///////////////////////////////////////////////////////////////////////////////
// big_if_statement.hpp
//
// Concepts:
//  Expression is...
//    a type which has a static member function, 
//    bool evaluate(Args);
//
// Note:
//  big_if_statement can't touch Args type,
//  so you should choose whether referenced or not.
//

#ifndef KETCHUP_BIG_IF_STATEMENT_H
#define KETCHUP_BIG_IF_STATEMENT_H

#ifdef _MSC_VER
#pragma once
#endif

#include <boost/concept_check.hpp>
#include <boost/mpl/reverse_fold.hpp>

namespace ketchup {

	namespace detail {

	///////////////////////////////////////////////////////////////////////////////
	// Note:
	//  If IfStatementSeq has the same IfStatement, VC7.1 duplicates
	//  the static member function, IfStatement::evaluate (LNK1179 error).
	//  I can't understand the problem,
	//  but VC7.1 seems to hate static member functions.
	//  So I define fold_op_workaround and...
	//
	// evaluate_condition
	//
	template< class Expression, class Args >
	typename Expression::type
	evaluate_expression(Args xs)
	{
		return Expression::evaluate(xs);
	}

	} // namespace detail
	
///////////////////////////////////////////////////////////////////////////////
// helper metafunctions
//
template< class ConditionExpression, class ReturnedExpression >
struct gen_if_statement
{
	typedef typename ConditionExpression condition_expression_type;
	typedef typename ReturnedExpression returned_expression_type;
};

template< class IfStatement >
struct get_condition_expression
{
	typedef typename IfStatement::condition_expression_type type;
};

template< class IfStatement >
struct get_returned_expression
{
	typedef typename IfStatement::returned_expression_type type;
};

///////////////////////////////////////////////////////////////////////////////
// if_statement_seq
//
template< class IfStatements, class Args, class FinalExpression >
struct if_statement_seq
{
	typedef typename FinalExpression::type result_type;

	struct fold_init_state
	{
		static result_type evaluate(Args xs)
		{
			boost::ignore_unused_variable_warning(xs);
			return detail::evaluate_expression<FinalExpression>(xs);
		}
	};
	
	// The placeholders say mpl_assert_not_arg, so I define
	struct fold_op_workaround
	{
		template< class State, class IfStatement >
		struct apply
		{
			typedef typename get_condition_expression<IfStatement>::type cond_expr_t;
			typedef typename get_returned_expression<IfStatement>::type ret_expr_t;

			static result_type evaluate(Args xs)
			{
				if (detail::evaluate_expression<cond_expr_t, Args>(xs))
					return detail::evaluate_expression<ret_expr_t, Args>(xs);
				else
					return State::evaluate(xs);
			}

			typedef apply type;
		};
	};

	typedef typename boost::mpl::reverse_fold<IfStatements,
		fold_init_state,
		// fold_op<boost::mpl::_1, boost::mpl::_2> // makes mpl_assert_not_arg
		fold_op_workaround
	>::type type;
};

} // namespace ketchup

#endif
