///////////////////////////////////////////////////////////////////////////////
// big_if_statement.hpp
//
// Concepts:
//  ConditionExpressionSeq is...
//    a MPL Sequence of ConditionExpression
//  ConditionExpression is...
//    a type which has a static member function, 
//    bool evaluate(Args);
//
// Note:
//  big_if_statement can't touch Args type,
//  so you should choose whether referenced or not.
//

#ifndef KETCHUP_BIG_IF_STATEMENT_H
#define KETCHUP_BIG_IF_STATEMENT_H

#ifdef _MSC_VER
#pragma once
#endif

#include <boost/concept_check.hpp>
#include <boost/mpl/reverse_fold.hpp>

namespace ketchup {	namespace detail {

///////////////////////////////////////////////////////////////////////////////
// Note:
//  If ConditionExpressionSeq has the same ConditionExpression, VC7.1 duplicates
//  the static member function, ConditionExpression::evaluate (LNK1179 error).
//  I can't understand the problem,
//  but VC7.1 seems to hate static member functions.
//  So I define fold_op_workaround and...
//
// evaluate_condition_expression
//
template< class ConditionExpression, class Args >
bool evaluate_condition_expression(Args xs)
{
	return ConditionExpression::evaluate(xs);
}

///////////////////////////////////////////////////////////////////////////////
// big_condition_expression
//
template< class ConditionExpressionSeq, class Args >
struct big_if_statement
{
	struct fold_init_state
	{
		static BOOL compile(Args xs)
		{
			boost::ignore_unused_variable_warning(xs);
			return FALSE;
		}
	};
	
	/*
	template< class State, class ConditionExpression >
	struct fold_op
	{
		static BOOL compile(Args xs)
		{
			if (detail::evaluate_condition_expression<IfStatement, Args>(xs))
				return TRUE;
			else
				return State::compile(xs);
		}
	};
	*/

	// The placeholders say mpl_assert_not_arg, so I define
	struct fold_op_workaround
	{
		template< class State, class ConditionExpression >
		struct apply
		{
			static BOOL compile(Args xs)
			{
				if (detail::evaluate_condition_expression<ConditionExpression, Args>(xs))
					return TRUE;
				else
					return State::compile(xs);
			}

			typedef apply type;
		};
	};

	typedef typename boost::mpl::reverse_fold<ConditionExpressionSeq,
		fold_init_state,
		// fold_op<boost::mpl::_1, boost::mpl::_2> // makes mpl_assert_not_arg
		fold_op_workaround
	>::type type;
};

} } // namespace ketchup::detail

#endif
